
#include "fastbf.h"
#include <stdint.h>
#include <stdio.h>

/*
simon@gamerarch:~/W/fastbf$ time ./bf < bench.bf
OK
real    0m0,563s
user    0m0,558s
sys     0m0,003s
simon@gamerarch:~/W/fastbf$ time ./bf < bench.bf
OK
real    0m0,562s
user    0m0,562s
sys     0m0,000s
simon@gamerarch:~/W/fastbf$ time ./bf < bench.bf
OK
real    0m0,569s
user    0m0,569s
sys     0m0,001s
*/

typedef struct {
    uint_fast16_t begin;
    uint_fast16_t depth;
} LoopEntry;

typedef struct {
    const char* program;
    size_t length;
    uint_fast16_t pc;
    uint_fast16_t sp;
    uint_fast16_t loop_depth;
    uint_fast16_t loops_length;
    uint_fast16_t loops_progress;
    LoopEntry loops[1000];
    uint8_t memory[30000];
} Context;

void v2_incr(Context* ctx) { ctx->memory[ctx->sp]++; }
void v2_decr(Context* ctx) { ctx->memory[ctx->sp]--; }
void v2_right(Context* ctx) { ctx->sp++; }
void v2_left(Context* ctx) { ctx->sp--; }
void v2_put(Context* ctx) { putc(ctx->memory[ctx->sp], stdout); }
void v2_get(Context* ctx) { ctx->memory[ctx->sp] = (uint8_t)getc(stdout); }

void v2_begin(Context* ctx)
{
    if (ctx->loops_progress < ctx->pc) {
        ctx->loops_progress++;
        ctx->loops[ctx->loops_length] = (LoopEntry) {
            .begin = ctx->pc,
            .depth = ctx->loop_depth,
        };
        ctx->loops_length++;
    }
    ctx->loop_depth++;
}

void v2_end(Context* ctx)
{
    if (ctx->memory[ctx->sp]) {
        int entry_index = 0;
        for (int i = (int)ctx->loops_length - 1; i >= 0; i--)
            if (ctx->loops[i].depth == ctx->loop_depth)
                entry_index = i;
        ctx->pc = ctx->loops[entry_index].begin;
    } else {
        ctx->loop_depth--;
    }
}

void v2_run_instruction(Context* ctx)
{
    printf(
        "pc = %4ld, sp = %4ld, ins = %c\n",
        ctx->pc,
        ctx->sp,
        ctx->program[ctx->pc]
    );
    switch (ctx->program[ctx->pc]) {
        case '+':
            v2_incr(ctx);
            break;
        case '-':
            v2_decr(ctx);
            break;
        case '>':
            v2_right(ctx);
            break;
        case '<':
            v2_left(ctx);
            break;
        case '.':
            v2_put(ctx);
            break;
        case ',':
            v2_get(ctx);
            break;
        case '[':
            v2_begin(ctx);
            break;
        case ']':
            v2_end(ctx);
            break;
    }
}

void v2(const char* program, size_t length)
{
    (void)length;
    Context ctx = {
        .program = program,
        .length = length,
        .pc = 0,
        .sp = 0,
        .loop_depth = 0,
        .loops_length = 0,
        .loops_progress = 0,
    };
    for (; ctx.pc < length; ctx.pc++) {
        v2_run_instruction(&ctx);
    }
}
